@@ -216,10 +216,10 @@ g_vfs_ftp_dir_cache_lookup_file_internal (GVfsFtpDirCache *  cache,
       g_vfs_ftp_dir_cache_entry_unref (entry);
     }
 
-  if (!g_vfs_ftp_task_is_in_error (task))
-    info = cache->funcs->lookup_uncached (task, file);
+  if (g_vfs_ftp_task_is_in_error (task))
+    return NULL;
 
-  return info;
+  return cache->funcs->lookup_uncached (task, file);
 }
 
 static GFileInfo *
@@ -241,6 +241,7 @@ g_vfs_ftp_dir_cache_resolve_symlink (GVfsFtpDirCache *  cache,
   GFileInfo *info, *result;
   GVfsFtpFile *tmp, *link;
   guint i, lookups = 0;
+  const char *target;
 
   if (!g_file_info_get_is_symlink (original) ||
       g_vfs_ftp_task_is_in_error (task))
@@ -248,10 +249,18 @@ g_vfs_ftp_dir_cache_resolve_symlink (GVfsFtpDirCache *  cache,
 
   info = g_object_ref (original);
   link = g_vfs_ftp_file_copy (file);
-  do
-    {
-      /* This must not happen, as we use one of our own GFileInfos */
-      g_assert (g_file_info_get_symlink_target (info) != NULL);
+  do {
+      target = g_file_info_get_symlink_target (info);
+      if (target == NULL)
+        {
+          /* This happens when bad servers don't report a symlink target.
+           * We now want to figure out if this is a directory or regular file,
+           * so we can at least report something useful.
+           */
+          g_object_unref (info);
+          info = cache->funcs->lookup_uncached (task, file);
+          break;
+        }
       tmp = link;
       link = cache->funcs->resolve_symlink (task, tmp, g_file_info_get_symlink_target (info));
       g_vfs_ftp_file_free (tmp);
@@ -508,6 +517,7 @@ g_vfs_ftp_dir_cache_funcs_process (GInputStream *        stream,
   int type;
   GVfsFtpFile *file;
   char *line, *s;
+  gsize length;
 
   /* protect against code reorg - in current code, error never is NULL */
   g_assert (error != NULL);
@@ -516,11 +526,17 @@ g_vfs_ftp_dir_cache_funcs_process (GInputStream *        stream,
   data = g_data_input_stream_new (stream);
   /* we use LF only, because the mozilla code can handle lines ending in CR */
   g_data_input_stream_set_newline_type (data, G_DATA_STREAM_NEWLINE_TYPE_LF);
-  while ((line = g_data_input_stream_read_line (data, NULL, cancellable, error)))
+  while ((line = g_data_input_stream_read_line (data, &length, cancellable, error)))
     {
       struct list_result result = { 0, };
       GTimeVal tv = { 0, 0 };
 
+      /* strip trailing \r - ParseFTPList only removes it if the line ends in \r\n,
+       * but we stripped the \n already.
+       */
+      if (length > 0 && line[length - 1] == '\r')
+        line[--length] = '\0';
+
       g_debug ("<<%2d <<  %s\n", debug_id, line);
       type = ParseFTPList (line, &state, &result);
       if (type != 'd' && type != 'f' && type != 'l')
@@ -551,6 +567,7 @@ g_vfs_ftp_dir_cache_funcs_process (GInputStream *        stream,
       g_free (s);
       if (file == NULL)
         {
+          g_debug ("# invalid filename, skipping");
           g_free (line);
           continue;
         }
